-- Copyright (c) 2017-2020 Nicolas Mailhot <nim@fedoraproject.org>
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <https://www.gnu.org/licenses/>.
--
-- SPDX-License-Identifier: GPL-3.0-or-later

-- This file contains Lua code used in rpm macros needed to create and process
-- source rpm (srpm) Go (golang) packages.
--
-- The resulting code must not use any package except redhat-rpm-macros

local fedora = require "fedora.common"

-- Core of %go_rpmname
local function rpmname(go_module, compatid)
  -- lowercase and end with '/'
  local   go_name = string.lower(rpm.expand(go_module) .. "/")
  -- remove eventual protocol prefix
  go_name         = string.gsub(go_name, "^http(s?)://",         "")
  -- remove eventual .git suffix
  go_name         = string.gsub(go_name, "%.git/*",              "")
  -- remove eventual git. prefix
  go_name         = string.gsub(go_name, "^git%.",               "")
  -- remove FQDN root (.com, .org, etc)
  -- will also remove vanity FQDNs such as "tools"
  go_name         = string.gsub(go_name, "^([^/]+)%.([^%./]+)/", "%1/")
  -- add golang prefix
  go_name         = "golang-" .. go_name
  -- compat naming additions
  local compatid = string.lower(rpm.expand(compatid))
  if  (compatid ~= nil) and (compatid ~= "") then
    go_name       = "compat-" .. go_name .. "-" .. compatid
 end
  -- special-case x.y.z number-strings as that’s an exception in our naming
  -- guidelines
  repeat
    go_name, i    = string.gsub(go_name, "(%d)%.(%d)",           "%1:%2")
  until i == 0
  -- replace various separators rpm does not like with -
  go_name         = string.gsub(go_name, "[%._/%-]+",            "-")
  -- because of the Azure sdk
  go_name         = string.gsub(go_name, "%-for%-go%-",          "-")
  -- Tokenize along - separators and remove duplicates to avoid
  -- golang-foo-foo-bar-foo names
  local result = ""
  local tokens = {}
  tokens["go"]     = true
  for token in string.gmatch(go_name, "[^%-]+") do
     if not tokens[token] then
        result = result .. "-" .. token
        tokens[token] = true
     end
  end
  -- reassemble the string, restore x.y.z runs, convert the vx.y.z
  -- Go convention to x.y.z as prefered in rpm naming
  result = string.gsub(result, "^-", "")
  result = string.gsub(result, ":", ".")
  -- some projects have a name that end up in a number, and *also* add release
  -- numbers on top of it, keep a - prefix before version strings
  result = string.gsub(result, "%-v([%.%d])", "-%1")
  return(result)
end

-- The rpm Go variables, split by initialisation stage:
local vars = {
  -- key variables that may control a declaration block
  key         = {"go_module",   "go_url",   "go_forge_suffix"},
  -- variables, that may fallback to suffixless variables
  shared      = {"go_license",  "go_description",
                 "go_licenses", "go_licenses_exclude",
                 "go_docs",     "go_docs_exclude"},
  -- variables, that may fallback to forge variables using %{go_forge_suffix}
  forge       = {"go_tag", "go_commit", "go_time", "go_version",
                 "go_extractdir"},
  forge_forge = {"go_patchlist"},
  -- other computed variables
  computed    = {"go_version", "go_version_core", "go_pre",
                 "go_filters", "go_license",
    "go_name",         "go_module_name",        "go_devel_name",
    "go_summary",      "go_module_summary",     "go_devel_summary",
    "go_description",  "go_module_description", "go_devel_description",
    "go_staging_list", "go_module_list",        "go_devel_list" },
  -- other variables
  other       = {      "go_module_header",      "go_devel_header"}
}

-- Return all the suffixes for which one of the Go key variables is set
-- Create a %{forgeurl} fallback if none exist yet
local function suffixes(verbose)
  local S = fedora.allsuffixes(vars["key"])
  if next(S) == nil then
    fedora.safeset("go_forge_suffix0", "0", verbose)
    fedora.safeset("go_url0", "%{forge_url0}", verbose)
    fedora.zalias({"go_url0"})
    S = {"0"}
  end
  return S
end

-- Executes the f function in a loop
--   – the loop is controled by index (a string):
--     — if the index value given to floop is nil, loop over the whole
--       range. Use readflag() to process rpm flag arguments safely.
--     — otherwise execute f for the specified index value only.
--   – f is passed index then otherargs as arguments.
--   – the index range is controlled by suffixes()
local function floop(f, index, otherargs, verbose)
  fedora.floop(f, index, suffixes(verbose), otherargs)
end

-- Core of %go_init
local function init(suffix, verbose, informative)
  local  forge = require "fedora.srpm.forge"
  local ismain = (suffix == "") or (suffix == "0")
  local vars_all = {}
  for k, _ in pairs(vars) do
    vars_all = fedora.mergelists({vars_all, vars[k]})
  end
  vars["all"] = vars_all
  if ismain then
    fedora.zalias(vars["all"],    verbose)
  else
    fedora.zalias(vars["shared"], verbose)
  end
  if not ismain then
    for _, v in ipairs(vars["shared"]) do
      if fedora.read(v) then
        fedora.safeset(v .. suffix, "%{" .. v .. "}", verbose)
      end
    end
  end
  local go_module_f = fedora.read("go_url" .. suffix)
  if go_module_f then
        go_module_f = string.gsub(go_module_f, "^http(s?)://", "")
        go_module_f = string.gsub(go_module_f, "/+$",          "")
    fedora.safeset("go_module" .. suffix, go_module_f, verbose)
  end
  local forge_suffix = fedora.read("go_forge_suffix" .. suffix)
  if (not forge_suffix) then
    local    go_url_f = fedora.read("go_module" .. suffix)
    go_url_f = "https://" .. go_url_f .. "/"
    fedora.safeset("go_url" .. suffix, go_url_f, verbose)
    go_url_f = fedora.read("go_url" .. suffix)
    forge_suffix = fedora.suffix("forge_url", go_url_f, true, verbose)
    fedora.set("go_forge_suffix" .. suffix, forge_suffix, verbose)
  end
  for _, v in ipairs(vars["forge_forge"]) do
    local fv = string.gsub(v, "^go_", "forge_") .. forge_suffix
    local gv = v .. suffix
    if fedora.read(v) then
      fedora.safeset(fv, "%{" .. gv .. "}", verbose)
    elseif fedora.read(fv) then
      fedora.safeset(gv, "%{" .. fv .. "}", verbose)
    end
  end
  for _, v in ipairs(vars["forge"]) do
    local fv = string.gsub(v, "^go_", "") .. forge_suffix
    local gv = v .. suffix
    if fedora.read(v) then
      fedora.safeset(fv, "%{" .. gv .. "}", verbose)
    elseif fedora.read(fv) then
      fedora.safeset(gv, "%{" .. fv .. "}", verbose)
    end
  end
  forge.init(forge_suffix, verbose, informative, false)
  local go_tag = fedora.read("go_tag" .. suffix)
  if go_tag then
    local tag_p = { version = "^v(%d+%.?%d+%.?%d+)",
                        pre = "^v%d+%.?%d*%.?%d*-([^-#]+)" }
    for v, p in tag_p do
      r = string.match(go_tag, p)
      if (r ~= nil) then
        fedora.safeset("go_" .. v .. suffix, r, verbose)
      end
    end
  end
  fedora.safeset("go_version"             .. suffix,
                 "0",
                 verbose)
  fedora.safeset("go_version_core"        .. suffix,
                 "%{go_version"           .. suffix .. "}",
                 verbose)
  fedora.safeset("go_filters"            .. suffix,
                 "%{?go_filters_default}",
                 verbose)
  fedora.safeset("go_name"               .. suffix,
                 rpmname("%{go_module"   .. suffix .. "}",
                         "%{?go_cid"     .. suffix .. "}"),
                 verbose)
  fedora.safeset("go_summary"            .. suffix,
                 "The %{go_name"         .. suffix .. "} Go software project",
                 verbose)
  fedora.safeset("go_description"        .. suffix,
                 "%{go_summary}.",
                 verbose)
  fedora.safeset("go_staging_list"        .. suffix,
                 "%{go_workdir}/%{go_name" .. suffix .. "}.tmp.lst",
                 verbose)
  local go_name = fedora.read("go_name"  .. suffix)
  fedora.safeset("go_module_name"        .. suffix,
                 go_name:gsub("-module$", "") .. "-module",
                 verbose)
  fedora.safeset("go_module_summary"     .. suffix,
                 "The %{go_module"       .. suffix .. "} Go module",
                 verbose)
  fedora.safeset("go_module_description" .. suffix,
                 [[%{expand:
%{?go_description]] .. suffix ..[[}

This package provides %{go_module]] .. suffix .. [[} Go module source code.
}]],
                 verbose)
  fedora.safeset("go_module_list"        .. suffix,
                 "%{go_workdir}/%{go_module_name" .. suffix .. "}.lst",
                 verbose)
  fedora.safeset("go_devel_name"          .. suffix,
                 string.gsub(go_name, "-devel$", "") .. "-devel",
                 verbose)
  fedora.safeset("go_devel_summary"       .. suffix,
                 "The %{go_module"        .. suffix .. "} GOPATH source code",
                 verbose)
  fedora.safeset("go_devel_description"   .. suffix,
                 [[%{expand:
%{?go_description]] .. suffix .. [[}

This package provides %{go_module]] .. suffix .. [[} legacy GOPATH source code.
}]],
                 verbose)
  fedora.safeset("go_devel_list"          .. suffix,
                 "%{go_workdir}/%{go_devel_name" .. suffix .. "}.lst",
                 verbose)
  for _, v in ipairs({"name", "summary", "version", "url", "license",
                      "description"}) do
    fedora.safeset("source_" .. v, "%{" .. "go_" .. v .. suffix .. "}",
                   verbose)
  end
  if ismain then
    fedora.zalias(fedora.mergelists({vars["key"], vars["forge"],
                                    vars["computed"]}), verbose)
  end
  -- Final spec variable summary if the macro was called with -i
  if informative then
    fedora.echo("Packaging variables read or set by %%go_init")
    fedora.echovars(vars["all"], suffix)
  end
end

-- Make currentgo* variables point to a specific suffix-indexed variable set
local function env(suffix, verbose)
  fedora.set_current(vars["all"], suffix, verbose)
  fedora.set_verbose(verbose)
end

-- Create a single package header for a golang-*-module subpackage
local function module_pkg()
  print(rpm.expand([[
%new_package -n %{__current_go_module_name}
Summary:        %{__current_go_module_summary}
Version:        %{__current_go_version}
License:        %{__current_go_license}
BuildRequires:  go-rpm-macros
Requires:       go-filesystem
%{?__current_go_module_header}
%description -n %{__current_go_module_name}
%wordwrap    -v   __current_go_module_description
]]))
end

-- Create a single package header for a golang-*-devel subpackage
local function devel_pkg()
  print(rpm.expand([[
%new_package -n %{__current_go_devel_name}
Summary:        %{__current_go_devel_summary}
Version:        %{__current_go_version}
License:        %{__current_go_license}
BuildRequires:  go-rpm-macros
Requires:       go-filesystem
%{?__current_go_devel_header}
%description -n %{__current_go_devel_name}
%wordwrap    -v   __current_go_devel_description
]]))
end

-- Core of %go_pkg
local function pkg(suffix, verbose)
  init(suffix, verbose, false)
  env( suffix, verbose)
  module_pkg()
  devel_pkg()
end

return {
  rpmname         = rpmname,
  suffixes        = suffixes,
  floop           = floop,
  init            = init,
  env             = env,
  pkg             = pkg,
}
